
import pandas as pd
import streamlit as st

from minitrade.utils.mtdb import MTDB

st.set_page_config(page_title='Settings', layout='wide')

log = st.sidebar.radio('Logs', [
    'TraderLog',
    'BacktestLog',
    'OrderValidatorLog',
    'ManualTrade',
    'IbOrderLog',
    'IbOrder',
    'IbTrade',
    'RawOrder',
    'Ticker',
])

data = MTDB.get_all(log, cls=dict)
data = [{k: str(v) if isinstance(v, pd.DataFrame) or v else v for k, v in x.items()} for x in data]

st.subheader(log)
st.write(pd.DataFrame(data))




from datetime import datetime, timezone

import akshare as ak
import pandas as pd

from minitrade.datasource.base import QuoteSource

# Need to manually copy libmini_racer.dylib on Mac M1, see https://github.com/sqreen/PyMiniRacer/issues/143


class EastMoneyQuoteSource(QuoteSource):
    '''EastMoney data source'''

    def _daily_bar(self, ticker, start, end):
        df: pd.DataFrame = ak.stock_zh_a_hist(
            symbol=ticker, period="daily",
            start_date=start.replace('-', ''),
            end_date=end.replace('-', '') if end else '20500101',   # magic number used in ak lib
            adjust='qfq')
        df = df.rename(columns={'日期': 'dt', '开盘': 'Open', '收盘': 'Close', '最高': 'High', '最低': 'Low', '成交量': 'Volume'})
        df['dt'] = pd.to_datetime(df['dt'], format="%Y-%m-%d")
        df = df.set_index('dt')
        df = df.tz_localize('Asia/Shanghai')
        df = df[['Open', 'High', 'Low', 'Close', 'Volume']]
        return df

    def _spot(self, tickers):
        try:
            src = [ak.stock_zh_a_spot_em, ak.fund_etf_spot_em, ak.fund_lof_spot_em]
            df = [f().rename(columns={'代码': 'ticker', '最新价': 'price'}).set_index('ticker') for f in src]
            s = pd.concat(df)['price']
            data = {ticker: s[ticker] if ticker in s.index else None for ticker in tickers}
            return pd.Series(data, name=datetime.now(timezone.utc)).astype(float)
        except Exception as e:
            raise AttributeError(f'Data error') from e





from minitrade.broker.base import Broker, BrokerAccount
from minitrade.datasource.base import QuoteSource
from minitrade.trader.trader import TradePlan


class InteractiveBrokersQuoteSource(QuoteSource):
    '''InteractiveBrokers data source'''

    def __init__(self, plan: TradePlan | str) -> None:
        super().__init__()
        self.plan = TradePlan.get_plan(plan) if isinstance(plan, str) else plan
        self.account = BrokerAccount.get_account(plan)
        self.broker = Broker.get_broker(self.account)

    def _daily_bar(self, ticker, start, end):
        self.broker.connect()
        return self.broker.daily_bar(self.plan, ticker, start, end)

    def _spot(self, tickers):
        self.broker.connect()
        return self.broker.spot(self.plan, tickers)



from __future__ import annotations

import io
import logging
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import Any
from urllib import request
from zoneinfo import ZoneInfo

import pandas as pd
import pandas_market_calendars as mcal
from tqdm.auto import tqdm

from minitrade.utils.mtdb import MTDB

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

__all__ = [
    'QuoteSource',
    'download_tickers',
]


class QuoteSource(ABC):
    '''
    QuoteSource is a base class that returns quote data for instruments. Extend this class
    to add a concrete implementation to get data from particular data source.
    '''

    AVAILABLE_SOURCES = ['Yahoo', 'EastMoney']
    '''A list of names for supported quote sources as input to `QuoteSource.get_source()`.'''

    @staticmethod
    def get_source(name: str, **kwargs: dict[str, Any]) -> QuoteSource:
        '''Get quote source by name

        Args:
            name: Quote source name
            kwargs: Keyword arguments to be passed to the quote source constructor

        Returns:
            A QuoteSource instance

        Raises:
            AttributeError: If the asked data source is not supported
        '''
        if name == 'Yahoo':
            from .yahoo import YahooQuoteSource
            return YahooQuoteSource(**kwargs)
        elif name == 'EastMoney':
            from .eastmoney import EastMoneyQuoteSource
            return EastMoneyQuoteSource()
        elif name == 'IB':
            from .ib import InteractiveBrokersQuoteSource
            return InteractiveBrokersQuoteSource(**kwargs)
        else:
            raise AttributeError(f'Quote source {name} is not supported')

    def today(self, ticker: str) -> datetime:
        '''Get today's date in a ticker's local timezone.'''
        tk = MTDB.get_one('Ticker', 'ticker', ticker)
        dt = datetime.now(ZoneInfo(tk['timezone'])).replace(hour=0, minute=0, second=0, microsecond=0)
        return dt

    def is_trading_now(self, ticker: str) -> bool:
        '''Check if a ticker is trading now.'''
        tk = MTDB.get_one('Ticker', 'ticker', ticker)
        calendar = mcal.get_calendar(tk['calendar'])
        today = datetime.now(ZoneInfo(tk['timezone']))
        schedule = calendar.schedule(
            start_date=(today - timedelta(days=30)).strftime('%Y-%m-%d'),
            end_date=(today + timedelta(days=30)).strftime('%Y-%m-%d')
        )
        return calendar.is_open_now(schedule, only_rth=True)

    @abstractmethod
    def _spot(self, tickers: list[str]) -> pd.Series:
        raise NotImplementedError()

    def spot(self, tickers: list[str] | str) -> pd.Series:
        '''Read current quote for a list of `tickers`.

        Args:
            tickers: Tickers as a list of string or a comma separated string without space

        Returns:
            Current quotes indexed by ticker as a pandas Series
        '''
        if isinstance(tickers, str):
            tickers = tickers.split(',')
        return self._spot(tickers)

    @abstractmethod
    def _daily_bar(self, ticker: str, start: str, end: str = None) -> pd.DataFrame:
        '''Same as `daily_bar()` for only one ticker.

        This should be overridden in subclass to provide an implemention.

        Returns:
            A dataframe with columns 'Open', 'High', 'Low', 'Close', 'Volume' indexed by datetime
        '''
        raise NotImplementedError()

    def daily_bar(
            self, tickers: list[str] | str,
            start: str = '2020-01-01', end: str = None, align: bool = True, normalize: bool = False) -> pd.DataFrame:
        '''Read end-of-day OHLCV data for a list of `tickers` starting from `start` date and ending on `end` date (both inclusive).

        Args:
            tickers: Tickers as a list of string or a comma separated string without space
            start: Start date in string format 'YYYY-MM-DD'
            end: End date in string format 'YYYY-MM-DD'
            align: True to align data to start on the same date, i.e. drop leading days when not all tickers have data available.
            normalize: True to normalize the close price on the start date to 1 for all tickers and scale all price data accordingly.

        Returns:
            A dataframe with 2-level columns, first level being the tickers, and the second level being columns 'Open', 'High', 'Low', 'Close', 'Volume'. The dataframe is indexed by datetime.
        '''
        try:
            if isinstance(tickers, str):
                tickers = tickers.split(',')
            data = {ticker: self._daily_bar(ticker, start, end) for ticker in tickers}
            df = pd.concat(data, axis=1).fillna(method='ffill')
            if align:
                start_index = df[df.notna().all(axis=1)].index[0]
                df = df.loc[start_index:, :]
                if normalize:
                    ohlc = ['Open', 'High', 'Low', 'Close']
                    for s in tickers:
                        df.loc[:, (s, ohlc)] = df.loc[:, (s, ohlc)] / df[s].loc[start_index, 'Close']
            return df
        except Exception:
            raise AttributeError(f'Data error')


def download_tickers():
    import warnings
    warnings.simplefilter("ignore")     # ignore warnings from pandas read_excel

    def download_nasdaq_traded_symbols():
        df = pd.read_csv('ftp://ftp.nasdaqtrader.com/SymbolDirectory/nasdaqtraded.txt',
                         sep='|', skipfooter=2, engine='python')
        df = df[df['Test Issue'] == 'N']
        df.to_sql('NasdaqTraded', MTDB.conn(), if_exists='replace', index=False)

    def download_szse_traded_symbols():
        df = pd.read_excel(
            'http://www.szse.cn/api/report/ShowReport?SHOWTYPE=xlsx&CATALOGID=1110&TABKEY=tab1',
            dtype={'A股代码': str, 'B股代码': str})
        df.to_sql('SzseTraded', MTDB.conn(), if_exists='replace', index=False)

    def download_szse_traded_etfs():
        df = pd.read_excel(
            'http://www.szse.cn/api/report/ShowReport?SHOWTYPE=xlsx&CATALOGID=1105&TABKEY=tab1&random=0.2921736379452743',
            dtype={'基金代码': str})
        df.to_sql('SzseTradedEtf', MTDB.conn(), if_exists='replace', index=False)

    def download_shse_traded_symbols():
        sse_stock_list_url = 'http://query.sse.com.cn//sseQuery/commonExcelDd.do?sqlId=COMMON_SSE_CP_GPJCTPZ_GPLB_GP_L&type=inParams&CSRC_CODE=&STOCK_CODE=&REG_PROVINCE=&STOCK_TYPE=1&COMPANY_STATUS=2,4,5,7,8'
        request_headers = {
            'X-Requested-With': 'XMLHttpRequest',
            'User-Agent':
            'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) '
            'Chrome/56.0.2924.87 Safari/537.36', 'Referer': 'http://www.sse.com.cn/assortment/stock/list/share/'}
        req = request.Request(sse_stock_list_url, headers=request_headers)
        result = request.urlopen(req).read()
        df = pd.read_excel(io.BytesIO(result), dtype={'A股代码': str, 'B股代码': str})
        df.to_sql('ShseTraded', MTDB.conn(), if_exists='replace', index=False)

    def download_hkse_traded_symbols():
        df = pd.read_excel(
            'https://www.hkex.com.hk/eng/services/trading/securities/securitieslists/ListOfSecurities.xlsx', skiprows=2,
            dtype={'Stock Code': str})
        df.to_sql('HkseTraded', MTDB.conn(), if_exists='replace', index=False)

    steps = [
        (download_nasdaq_traded_symbols, 'select Symbol as ticker, "Security Name" as name from NasdaqTraded', 'NASDAQ', 'America/New_York', ''),
        (download_shse_traded_symbols, 'select A股代码 as ticker, 证券简称 as name from ShseTraded', 'SSE', 'Asia/Shanghai', 'SH'),
        (download_szse_traded_symbols, 'select A股代码 as ticker, A股简称 as name from SzseTraded', 'SSE', 'Asia/Shanghai', 'SZ'),
        (download_szse_traded_etfs, 'select 基金代码 as ticker, 基金简称 as name from SzseTradedEtf', 'SSE', 'Asia/Shanghai', 'SZ'),
        (download_hkse_traded_symbols, 'select "Stock Code" as ticker, "Name of Securities" as name from HkseTraded', 'HKEX', 'Asia/Hongkong', 'HK'),
    ]
    MTDB.conn().execute('drop table if exists Ticker')
    t = tqdm(steps, total=len(steps), unit="step", desc='Downloading stock symbols', leave=False)
    for downloader, sql, calendar, timezone, yahoo_modifier in t:
        downloader()
        df = pd.read_sql(sql, MTDB.conn())
        df['calendar'] = calendar
        df['timezone'] = timezone
        df['yahoo_modifier'] = yahoo_modifier
        df.to_sql('Ticker', MTDB.conn(), if_exists='append', index=False)
    t.close()




from datetime import datetime, timedelta, timezone

import pandas as pd
import yfinance as yf

from minitrade.datasource import QuoteSource
from minitrade.utils.config import config


class YahooQuoteSource(QuoteSource):

    def __init__(self, proxy: str = None):
        ''' Yahoo data source

        Parameters
        ----------
        proxy: str
            Http proxy URI to override currently setting if not None
        '''
        self.proxy = proxy or config.sources.yahoo.proxy

    def _daily_bar(self, ticker, start, end):
        # Push 1 day out to include "end" in final data
        end_1 = end and (datetime.strptime(end, '%Y-%m-%d') + timedelta(days=1)).strftime('%Y-%m-%d')
        tk = yf.Ticker(ticker)
        df: pd.DataFrame = tk.history(start=start, end=end_1, interval='1d',
                                      auto_adjust=True, proxy=self.proxy, timeout=10)
        df = df[['Open', 'High', 'Low', 'Close', 'Volume']]
        # Today's data from history api are not reliable. Replace them with spot prices.
        today = self.today(ticker)
        if (not df.loc[today:].empty or not end or end >= today.strftime('%Y-%m-%d')) and self.is_trading_now(ticker):
            tz = df.index.tzinfo
            df = df[df.index < today]
            df.index = df.index.tz_convert(timezone.utc)
            row = dict(zip(df.columns, [
                tk.basic_info['open'],
                tk.basic_info['dayHigh'],
                tk.basic_info['dayLow'],
                tk.basic_info['lastPrice'],
                tk.basic_info['lastVolume']]))
            df = pd.concat([df, pd.DataFrame(row, index=[today.astimezone(timezone.utc)])])
            df.index = df.index.tz_convert(tz)
        return df

    def _spot(self, tickers):
        try:
            data = {ticker: yf.Ticker(ticker).basic_info['lastPrice']
                    if self.is_trading_now(ticker) else None for ticker in tickers}
            df = pd.Series(data, name=datetime.now(timezone.utc)).astype(float)
            return df
        except Exception as e:
            raise AttributeError(f'Data error') from e
